\AMaze3D2.inc	11-Aug-2009	Loren Blaney	loren.blaney@idcomm.com
\Displays randomly generated spherical mazes that can be manipulated in 3D
\Compile with 32-bit XPL0 (XPLPX, xpx.bat)
\
\ A right-hand coordinate system is used
\
\				|   Z  (into screen)
\				|  /
\				| /
\				|/
\			--------+-------- X
\			       /|
\			      / |
\			     /	|
\				|
\				Y

include	c:\cxpl\codes;		\intrinsic code definitions (including Paint)

def	VidMode=$118,		\video mode: 24-bit color graphics
	ScrW=1024, ScrH=768,	\screen width and height, in pixels
	ImW=ScrH, ImH=ScrH;	\image width and height in pixels (a square)

int	CpuReg;			\address of CPU register array (from GetReg)
int	Image(ImW*ImH);		\image buffer, gets copied to the graphic screen

def	Bel=$07, CR=$0D, Esc=$1B;				\control chars
def	UpArrow=$48, DnArrow=$50, LtArrow=$4B, RtArrow=$4D;	\scan codes



proc	Fatal(Msg);		\Display error message and abort program
char	Msg;
[SetVid(3);
Text(0, Msg);
exit;
];	\Fatal



func	WExtend(N);		\Sign-extend 16-bit word to 32-bit double word
int	N;
return if N & $8000 then N ! $FFFF0000 else N & $0000FFFF;



func	CallInt(Int, AX, BX, CX, DX, BP, DS, ES); \Call software interrupt
int	Int, AX, BX, CX, DX, BP, DS, ES; \(unused arguments need not be passed)
[CpuReg(0):= AX;
 CpuReg(1):= BX;
 CpuReg(2):= CX;
 CpuReg(3):= DX;
 CpuReg(6):= BP;
 CpuReg(9):= DS;
 CpuReg(11):= ES;
 SoftInt(Int);
 return WExtend(CpuReg(0));	\return value in AX register (as 32 bits)
];	\CallInt



func	GetKey;			\Get character from keyboard (wait if necessary)
int	SC, Ch;			\This is a low-level routine with no echo,
[				\ no Ctrl+C, and no cursor
SC:= CallInt($16, $0000);	\BIOS function $00
Ch:= SC & $FF;
if Ch = 0 then Ch:= -(SC>>8);	\return non-ASCII chars as negative scan code
return Ch;
];	\GetKey



func	OpenMouse;		\Initialize mouse; returns 'false' if it fails
[CallInt($21, $3533);		\Make sure mouse vector $33 points to something
if WExtend(CpuReg(1))=0 & WExtend(CpuReg(11))=0 then return false;
CallInt($33, $21);		\reset mouse for Windows
CallInt($33, $0B);		\read motion counters to reset them for Windows
return CallInt($33, $0000);	\h/w reset mouse and return status
];	\OpenMouse

\-------------------------------------------------------------------------------

proc	VectSub(V1, V2, V3);	\Subtract two 3D vectors
real	V1, V2, V3;		\V3:= V1 - V2
[V3(0):= V1(0) - V2(0);
 V3(1):= V1(1) - V2(1);
 V3(2):= V1(2) - V2(2);
];	\VectSub



proc	VectMul(V1, S, V2);	\Multiply 3D vector by a scaler
real	V1, S, V2;		\V2:= V1 * S
[V2(0):= V1(0)*S;
 V2(1):= V1(1)*S;
 V2(2):= V1(2)*S;
];	\VectMul



func real VectMag(V);		\Return the magnitude (length) of a 3D vector
real	V;
return sqrt(V(0)*V(0) + V(1)*V(1) + V(2)*V(2));



proc	Unitize(V);		\Convert a 3D vector to its unit vector
real	V;			\vector to convert
real	M;			\magnitude of vector
[M:= VectMag(V);
V(0):= V(0)/M;
V(1):= V(1)/M;
V(2):= V(2)/M;
];	\Unitize



func real DotProd(V1, V2);	\Return the dot product of two 3D vectors
real	V1, V2;			\V1  V2
return V1(0)*V2(0) + V1(1)*V2(1) + V1(2)*V2(2);



proc	CrossProd(V1, V2, V3);	\Calculate the cross product of two 3D vectors
real	V1, V2, V3;		\V3:= V1 x V2
[V3(0):= V1(1)*V2(2) - V1(2)*V2(1);
 V3(1):= V1(2)*V2(0) - V1(0)*V2(2);
 V3(2):= V1(0)*V2(1) - V1(1)*V2(0);
];	\CrossProd



proc	Rotate(V, R, P, W); \Rotate vector in 3 dimensions
real	V,		\3D vector, a 3-element array with X, Y and Z components
	R, P, W;	\Roll, Pitch and yaW (radians)
real	SW, SP, SR, CW, CP, CR, T;
[SR:= Sin(R);   SP:= Sin(P);   SW:= Sin(W);
 CR:= Cos(R);   CP:= Cos(P);   CW:= Cos(W);

\Rotate about X axis (Roll):
T:= V(1)*CR - V(2)*SR;		\		       	 
V(2):= V(2)*CR + V(1)*SR;	\		      1   0   0 
V(1):= T;			\ [X Y Z]:= [X Y Z] * 0   CR  SR
				\		      0  -SR  CR
				\		       	 
\Rotate about Y axis (Pitch):
T:= V(0)*CP + V(2)*SP;		\		       	 
V(2):= V(2)*CP - V(0)*SP;	\		      CP  0  -SP
V(0):= T;			\ [X Y Z]:= [X Y Z] * 0   1   0 
				\		      SP  0   CP
				\		       	 
\Rotate about Z axis (yaW):
T:= V(0)*CW - V(1)*SW;		\		       	  
V(1):= V(1)*CW + V(0)*SW;	\		       CW  SW	0 
V(0):= T;			\ [X Y Z]:= [X Y Z] * -SW  CW	0 
				\		       0   0	1 
];	\Rotate					       	  

\===============================================================================

proc	EraseImage;		\Erase the Image array
int	I;
begin
for I:= 0, ImW*ImH-1 do Image(I):= 0;
end;	\EraseImage



int	PenX, PenY;		\pen coordinates (pixels) for drawing onto Image

proc	MovePen(X, Y);		\Position drawing pen
int	X, Y;
[PenX:= X;  PenY:= Y];



proc	DrawLine(X, Y, C); \Draw line on Image from PenX,PenY to X,Y in color C
int	X, Y, C;	\coordinates of end-of-line, color
int	DX, DY,		\delta X, delta Y
	IX, IY,		\plotting increments along X and Y axes
	SY,		\flag: step along Y direction
	D,		\decision variable: move straight or diagonally
	E, NE,		\East, Northeast (straight or diagonal)
	T, I;		\scratch
[DX:= X - PenX;				\delta X
if DX >= 0 then				\if positive then increment
	IX:= 1
else	[DX:= -DX;   IX:= -1];		\else absolute value and decrement

DY:= Y - PenY;				\same for Y direction
if DY >= 0 then
	IY:= 1
else	[DY:= -DY;   IY:= -1];

if DX >= DY then			\step along X direction
	SY:= false
else	[SY:= true;			\step along Y direction
	T:= DX;   DX:= DY;   DY:= T;	\swap DX and DY
	];

D:= 2*DY - DX;				\set up decision variable stuff
E:= 2*DY;
NE:= 2*(DY-DX);

if PenX>=0 & PenX<ImW & PenY>=0 & PenY<ImH then	\clip
	Image(PenX + PenY*ImW):= C;	\plot initial point
for I:= 1, DX do
	[if D < 0 then
		[D:= D + E;
		if SY then PenY:= PenY + IY
		      else PenX:= PenX + IX;
		]
	else	[D:= D + NE;		\move diagonally
		PenX:= PenX + IX;
		PenY:= PenY + IY;
		];
	if PenX>=0 & PenX<ImW & PenY>=0 & PenY<ImH then	\clip
		Image(PenX + PenY*ImW):= C;
	];
];	\DrawLine

\-------------------------------------------------------------------------------

proc	DrawTriangle(Tri);	\Draw a filled triangle onto Image array
int	Tri;			\array of 3 points(x,y) and one 8-bit color
int	X0,Y0, X1,Y1, X2,Y2, Color;
int	X, Y, T;
int	Top, Bot,		\top and bottom Y coordinates
	YImW;			\Y coordinate times Image width
int	Edge1, Edge2;		\X coordinates along edges of triangles
real	REdge1, REdge2, Slope1, Slope2;
int	I;

	proc	DrawHalf;	\(optimized procedure with no local variables)
	[if Bot > ImH then Bot:= ImH-1;			\clip to bottom of Image
	YImW:= Top*ImW;
	for Y:= Top, Bot-1 do				\from top to bottom...
		[if Y >= 0 then				\clip to top of Image
			[Edge1:= fix(REdge1);  Edge2:= fix(REdge2);
			if Edge1 > Edge2 then		\order left to right
				[T:= Edge1;  Edge1:= Edge2;  Edge2:= T];
			if Edge1 < 0 then Edge1:= 0;	\clip to left side
			if Edge2 > ImW then Edge2:= ImW;\clip to right side
			for I:= Edge1+YImW, Edge2-1+YImW do
				Image(I):= Color;		\fill scanline
			];
		REdge1:= REdge1 + Slope1;		\adj X coords according
		REdge2:= REdge2 + Slope2;		\ to slopes
		YImW:= YImW + ImW;
		];
	];	\DrawHalf

[	\DrawTriangle
X0:= Tri(0,0);			\get vertices
X1:= Tri(1,0);
X2:= Tri(2,0);

if X0>=ImW & X1>=ImW & X2>=ImW then return;	\clip entire triangle
if X0<0 & X1<0 & X2<0 then return;

Y0:= Tri(0,1);
Y1:= Tri(1,1);
Y2:= Tri(2,1);

if Y0 > Y1 then			\sort vertices from top to bottom (Y0<=Y1<=Y2)
	[T:= Y1;  Y1:= Y0;  Y0:= T;
	 T:= X1;  X1:= X0;  X0:= T;
	];
if Y0 > Y2 then
	[T:= Y2;  Y2:= Y0;  Y0:= T;
	 T:= X2;  X2:= X0;  X0:= T;
	];
if Y0 >= ImH then return;	\clip
if Y1 > Y2 then
	[T:= Y2;  Y2:= Y1;  Y1:= T;
	 T:= X2;  X2:= X1;  X1:= T;
	];
if Y2 < 0 then return;		\clip

if Y1 # Y0 then Slope1:= float(X1-X0) / float(Y1-Y0);
if Y2 # Y0 then Slope2:= float(X2-X0) / float(Y2-Y0);
REdge1:= float(X0);  REdge2:= REdge1;
Top:= Y0;  Bot:= Y1;
Color:= Tri(3);
DrawHalf;

if Y2 # Y1 then Slope1:= float(X2-X1) / float(Y2-Y1);
REdge1:= float(X1);
Top:= Y1;  Bot:= Y2;
DrawHalf;
];	\DrawTriangle



proc	DrawQuad(Quad);		\Draw a filled quadrilateral onto Image array
int	Quad;			\array of 4 points(x,y) and one 8-bit color
int	Tri(4,2);
[Tri(3):= Quad(4);		\color

Tri(0,0):= Quad(0,0);
Tri(0,1):= Quad(0,1);
Tri(1,0):= Quad(1,0);
Tri(1,1):= Quad(1,1);
Tri(2,0):= Quad(2,0);
Tri(2,1):= Quad(2,1);
DrawTriangle(Tri);

Tri(0,0):= Quad(0,0);
Tri(0,1):= Quad(0,1);
Tri(1,0):= Quad(2,0);
Tri(1,1):= Quad(2,1);
Tri(2,0):= Quad(3,0);
Tri(2,1):= Quad(3,1);
DrawTriangle(Tri);
];	\DrawQuad
