\HexMaze.xpl	6-Jul-2009	Loren Blaney	loren.blaney@idcomm.com
\Generates a random hexagonal maze

cond false;
              x: 0_ 1_ 2_ 3_ 4_
                 __    __    __          __    __    __
            y:  /00\__/20\__/40\     ->    \__/  \__/  \
   5x3      _0  \__/10\__/30\__/        \__    __/     /
  Maze          /01\__/21\__/41\        /  \__   \  /  \
            _1  \__/11\__/31\__/        \      __/  \    ->
                /02\__/22\__/42\        /  \__/   __/  \
            _2  \__/12\__/32\__/        \__    __/   __/
                /  \__/  \__/  \        /  \__/  \__/  \
cond true;

include	c:\cxpl\codes;

def	Cols=17, Rows=11;	\dimensions of the maze (cells)

int	UpLeftWall(Cols+1,Rows),\boolean: upper left wall of cell is not opened
	DnLeftWall(Cols+1,Rows),\boolean: downward left wall of cell is unopened
				\ extra wall columns forms right border of maze
	Ceiling(Cols,Rows+1);	\boolean: ceiling of cell has not been opened
				\ extra ceiling row forms bottom border of maze


def	Size3=10, Size4=20, Size5=24;		\size of hexagons
def	Color=14;
int	XX, YY;			\current coordinates (pixels)

proc	Draw(Ch);		\Draw a character on graphic screen
int	Ch;
begin
case Ch of
  ^/:	[Move(XX, YY);   XX:= XX + Size3;   Line(XX, YY-Size4, Color)];
  ^\:	[Move(XX, YY-Size4);   XX:= XX + Size3;   Line(XX, YY, Color)];
  ^ :	XX:= XX + Size3;
  ^_:	[Move(XX, YY);   XX:= XX + Size5;   Line(XX, YY, Color)]
other	[];
end;	\Draw



proc	ShowMaze;		\Display the maze
int	X, Y;			\cell coordinates
begin
Clear;
XX:= 0;   YY:= 0;
for Y:= 0, Rows do
	begin
	for X:= 0, Cols-1 do
		begin
		if (X&1) = 0 then		\even column
			begin
			if Y>0 then
			   if DnLeftWall(X, Y-1) then
				Draw(^\)
			   else	Draw(^ )
			else Draw(^ );

			if Ceiling(X, Y) then
				Draw(^_)
			else	XX:= XX + Size5;
			end
		else	begin			\odd column
			if Y>0 then
			   if UpLeftWall(X, Y-1) then
				Draw(^/)
			   else	Draw(^ )
			else Draw(^ );

			XX:= XX + Size5;
			end;
		end;

	\Close off
	if Y>0 then
	    if (Cols&1) = 0 then		\even column
		[if DnLeftWall(Cols, Y-1) then Draw(^\)]
	    else				\odd column
		[if UpLeftWall(Cols, Y-1) then Draw(^/)];
	XX:= 0;   YY:= YY + Size4;

	for X:= 0, Cols-1 do
		begin
		if (X&1) = 0 then		\even column
			begin
			if Y<Rows then
			    if UpLeftWall(X, Y) then
				[Draw(^/);   XX:= XX + Size5]
			    else XX:= XX + Size5 + Size3
			else [Draw(^/);   XX:= XX + Size5]; \prevent wrap around
			end
		else	begin			\odd column
			if Y>0 then
			   if DnLeftWall(X, Y-1) then
				Draw(^\)
			   else	Draw(^ )
			else Draw(^\);		\prevent wrap around

			if Ceiling(X, Y) then
				Draw(^_)
			else	XX:= XX + Size5;
			end;
		end;
	\Close off
	if (Cols&1) = 0 then			\even column
		begin
		if Y<Rows & Y>0 then
		    if UpLeftWall(Cols, Y) then Draw(^/);
		end
	else					\odd column
		if Y>0 then
			[if DnLeftWall(Cols, Y-1) then Draw(^\)]
		else Draw(^\);			\prevent wrap around
	XX:= 0;   YY:= YY + Size4;
	end;
end;	\ShowMaze

\-------------------------------------------------------------------------------

proc	GenMaze;		\Generate the maze
int	NumConnected, X, Y, StartX, StartY;
int	UpOk, DnOk, LtUpOk, RtUpOk, LtDnOk, RtDnOk, Dir;
int	Connected(Cols,Rows);	\boolean: maze cell is connected to other cells
[for Y:= 0, Rows-1 do 		\all cells are initially disconnected
    for X:= 0, Cols-1 do
	Connected(X,Y):= false;

StartX:= Ran(Cols);  StartY:= Ran(Rows);	\pick a starting cell at random
Connected(StartX,StartY):= true;
NumConnected:= 1;

repeat	[X:= StartX;  Y:= StartY;
	loop	[\quit loop when all adjacent cells are connected
		UpOk:= false;    DnOk:= false;
		LtUpOk:= false;  RtUpOk:= false;
		LtDnOk:= false;  RtDnOk:= false;
		if Y-1 >= 0 then UpOk:= not Connected(X,Y-1);
		if Y+1 < Rows then DnOk:= not Connected(X,Y+1);
		if X-1 >= 0 then
		  if X&1 \odd column\ then
			LtUpOk:= not Connected(X-1, Y)	\same as moving left
		  else	if Y-1 >= 0 then LtUpOk:= not Connected(X-1, Y-1);
		if X+1 < Cols then
		  if X&1 \odd column\ then
			RtUpOk:= not Connected(X+1, Y)	\same as moving right
		  else	if Y-1 >= 0 then RtUpOk:= not Connected(X+1, Y-1);
		if X-1 >= 0 then
		  if X&1 \odd column\ then
			if Y+1 < Rows then LtDnOk:= not Connected(X-1, Y+1)
		  else	LtDnOk:= not Connected(X-1, Y);	\same as moving left
		if X+1 < Cols then
		  if X&1 \odd column\ then
			if Y+1 < Rows then RtDnOk:= not Connected(X+1, Y+1)
		  else	RtDnOk:= not Connected(X+1, Y);	\same as moving right
		if ~(UpOk ! DnOk ! LtUpOk ! RtUpOk ! LtDnOk ! RtDnOk) then quit;

		Dir:= Ran(6);			\choose direction randomly
		loop	[case Dir of		\ and try to connect that way
			  0: if UpOk then
				[Ceiling(X,Y):= false;	\remove ceiling
				Y:= Y-1;	\move into cell above
				quit;
				];
			  1: if DnOk then
				[Y:= Y+1;
				Ceiling(X,Y):= false;
				quit;
				];
			  2: if LtUpOk then
				[UpLeftWall(X,Y):= false;
				if (X&1) = 0 \even column\ then Y:= Y-1;
				X:= X-1;
				quit;
				];
			  3: if RtUpOk then
				[if (X&1) = 0 \even column\ then Y:= Y-1;
				X:= X+1;
				DnLeftWall(X,Y):= false;
				quit;
				];
			  4: if LtDnOk then
				[DnLeftWall(X,Y):= false;
				if X&1 \odd column\ then Y:= Y+1;
				X:= X-1;
				quit;
				];
			  5: if RtDnOk then
				[if X&1 \odd column\ then Y:= Y+1;
				X:= X+1;
				UpLeftWall(X,Y):= false;
				quit;
				]
			other	[];

			Dir:= Dir+1;		\try next direction
			if Dir > 5 then Dir:= 0;
			];

		Connected(X,Y):= true;		\connected a cell, so mark it
		NumConnected:= NumConnected+1;	\ and count it
		];	\loop

	repeat	[\move starting point to next connected cell
		if StartX+1 < Cols then StartX:= StartX+1
		else	[StartX:= 0;		\wrap-around raster pattern
			if StartY+1 < Rows then StartY:= StartY+1
			else StartY:= 0;
			];
		];
	until Connected(StartX,StartY);
	];
until	NumConnected >= Rows*Cols;		\all cells connected?
];	\GenMaze

\-------------------------------------------------------------------------------

int	X, Y;

[	\Main
for Y:= 0, Rows-1 do 
    for X:= 0, Cols do
	[UpLeftWall(X,Y):= true;			\all walls are erected
	 DnLeftWall(X,Y):= true;
	];
for Y:= 0, Rows do 
    for X:= 0, Cols-1 do
	Ceiling(X,Y):= true;			\all ceilings are in place
GenMaze;

UpLeftWall(0,    Ran(Rows)):= false;		\make entrance and exit doorways
UpLeftWall(Cols, Ran(Rows)):= false;

SetVid($12);
ShowMaze;
if ChIn(1) then [];				\wait for keystroke
SetVid(3);
];	\Main
