\MazePath.xpl	17-Oct-2011
\Maze Maker and Solver that Compares Depth-First Search to Breath-First Search
\Compile with 32-bit XPL0 (xpx.bat)

include c:\cxpl\codes;			\intrinsic routine 'code' declarations

def	Cols=48, Rows=48;		\dimensions of maze (cells)
def	X0 = (320-Cols*2*2)/2, Y0 = (200-Rows*2*2)/4;	\centers maze on screen
def	StartX=Cols/5, StartY=Rows/5,	\locations of start and goal cells
	GoalX=Cols*4/5, GoalY=Rows*4/5;
int	LeftWall(Cols+1, Rows+1),	\left wall for each cell + right border
	Ceiling(Cols+1, Rows+1),	\ceiling for each cell + bottom border
	Visited(Cols+1, Rows+1);	\(+1 for consistency and uniformity)

def	Black, Blue, Green, Cyan, Red, Magenta, Brown, White,  \attribute colors
	Gray, LBlue, LGreen, LCyan, LRed, LMagenta, Yellow, BWhite; \EGA palette
def	Bel=$07, Esc=$1B;		\ASCII control characters



proc	Exit;			\Make a clean exit from program
begin
SetVid($03);			\restore normal text mode
exit;
end;	\Exit



proc	Delay(TD);		\Time delay in microseconds
\This version works for horrible Windows XP
int	TD;	\(maximum is 26000, even for 32-bit XPL)
int	Cnt, T0;
begin
Cnt:= TD*1000/838;
T0:= GetTime;
repeat until ((GetTime-T0) & $FFFF) >= Cnt;  \25ms = 25e6ns; 25e6ns/838 = 29833
end;	\Delay

\-------------------------------------------------------------------------------

proc	BFS(X, Y);		\Solve maze using Breadth-First-Search
int	X, Y;			\cell coordinates
int	F, E,			\fill and empty indexes
	T,			\temporary scratch
	Q(Cols*Rows*2);		\queue (FIFO)


	proc	EnQ(X, Y);	\Enqueue coordinates
	int	X, Y;
	begin
	Q(F):= X;  F:= F+1;
	Q(F):= Y;  F:= F+1;
	end;	\EnQ


	proc	DeQ;		\Dequeue coordinates
	begin
	X:= Q(E);  E:= E+1;
	Y:= Q(E);  E:= E+1;
	end;	\DeQ


begin	\BFS
F:= 0;  E:= 0;
EnQ(X, Y);
loop	begin
	DeQ;
	Point(X*4+2+X0, Y*4+2+Y0, LRed);	\show what's happening
	Delay(3000);				\but don't go too fast

	if X=GoalX & Y=GoalY then		\solved!
		begin
		\show shortest path
		Point(X*4+2+X0, Y*4+2+Y0, LGreen);
		repeat	T:= Visited(X, Y);	       \trace path back to start
			Y:= T>>8;  X:= (T&$00FF) - 1;
			Point(X*4+2+X0, Y*4+2+Y0, LGreen);
			Delay(3000);		       \but don't go too fast
		until	X=StartX & Y=StartY;
		quit;				       \all done
		end;

	if ~LeftWall(X+1, Y) then if Visited(X+1, Y)=0 then
		[Visited(X+1, Y):= X+1 ! Y<<8;  EnQ(X+1, Y)];	\go right
	if ~Ceiling(X, Y+1) then if Visited(X, Y+1)=0 then
		[Visited(X, Y+1):= X+1 ! Y<<8;  EnQ(X, Y+1)];	\go down
	if ~LeftWall(X, Y) then if Visited(X-1, Y)=0 then
		[Visited(X-1, Y):= X+1 ! Y<<8;  EnQ(X-1, Y)];	\go left
	if ~Ceiling(X, Y) then if Visited(X, Y-1)=0 then
		[Visited(X, Y-1):= X+1 ! Y<<8;  EnQ(X, Y-1)];	\go up
	end;
end;	\BFS

\-------------------------------------------------------------------------------

int	Solved;			\flag to unwind recursion, once solved

proc	DFS(X, Y);		\Solve maze using Depth-First-Search
int	X, Y;
begin
if Solved ! Visited(X, Y) then return;
Visited(X, Y):= true;			\mark current cell as visited
Point(X*4+2+X0, Y*4+2+Y0, LGreen);	\show what's happening
Delay(4000);				\but don't go too fast

if X=GoalX & Y=GoalY then [Solved:= true;  return];

if ~LeftWall(X+1, Y) then DFS(X+1, Y);	\go right
if ~Ceiling(X, Y+1)  then DFS(X, Y+1);	\go down
if ~LeftWall(X, Y)   then DFS(X-1, Y);	\go left
if ~Ceiling(X, Y)    then DFS(X, Y-1);	\go up

if Solved then return;
Point(X*4+2+X0, Y*4+2+Y0, LRed);	\change dot to wrong-way red
Delay(4000);
end;	\DFS

\-------------------------------------------------------------------------------

proc	ShowMaze;		\Display the maze
int	X, Y;
begin
for Y:= 0 to Rows do
	begin
	for X:= 0 to Cols do
		begin
		Point(X*4+X0, Y*4+Y0, LBlue);
		if X#Cols & Ceiling(X, Y) then
			[Move(X*4+X0, Y*4+Y0);  Line(X*4+3+X0, Y*4+Y0, LBlue)];
		end;
	for X:= 0 to Cols do
		if Y#Rows & LeftWall(X, Y) then
			[Move(X*4+X0, Y*4+Y0);  Line(X*4+X0, Y*4+3+Y0, LBlue)];
	end;
Point(StartX*4+2+X0, StartY*4+2+Y0, Yellow);	\show start and goal locations
Point(GoalX*4+2+X0, GoalY*4+2+Y0, Yellow);
end;	\ShowMaze



proc	ConnectFrom(X, Y);	\Connect cells starting from cell X,Y
int	X, Y;
int	Dir, Dir0;
begin
Visited(X, Y):= true;			\mark current cell as connected
Dir:= Ran(4);				\randomly choose a direction
Dir0:= Dir;				\save this initial direction
repeat	case Dir of			\try to connect to cell at Dir
	  0: if X+1<Cols then if ~Visited(X+1, Y) then	    \go right
		[LeftWall(X+1, Y):= false; ConnectFrom(X+1, Y)];
	  1: if Y+1<Rows then if ~Visited(X, Y+1) then	    \go down
		[Ceiling(X, Y+1):= false;  ConnectFrom(X, Y+1)];
	  2: if X-1>=0 then if ~Visited(X-1, Y) then	    \go left
		[LeftWall(X, Y):= false;	 ConnectFrom(X-1, Y)];
	  3: if Y-1>=0 then if ~Visited(X, Y-1) then	    \go up
		[Ceiling(X, Y):= false;	 ConnectFrom(X, Y-1)]
	other	[];			\(never occurs)
	Dir:= Dir+1 & $03;		\next direction
until	Dir = Dir0;
end;	\ConnectFrom



proc	GenMaze;		\Make a randomly generated maze
int	X, Y;
begin
for Y:= 0 to Rows do
    for X:= 0 to Cols do
	begin
	LeftWall(X, Y):= true;		\start with all walls and
	Ceiling(X, Y):= true;		\ ceilings erected,
	Visited(X, Y):= false; 		\ and no cells visited for connection
	end;

ConnectFrom(Ran(Cols), Ran(Rows));	\make maze starting from a random cell

\Make some alternate pathways (so right-hand method won't work)
for X:= 0, Cols do
	LeftWall(Ran(Cols-1)+1, Ran(Rows-1)+1):= false;
for Y:= 0, Rows do
	Ceiling(Ran(Cols-1)+1, Ran(Rows-1)+1):= false;
end;	\GenMaze

\-------------------------------------------------------------------------------

int	Seed, X, Y;
begin	\Main
SetVid($13);				\320x200x8 graphics

Seed:= Ran(30000);
RanSeed(Seed);				\set random number generator's seed
GenMaze;
ShowMaze;
if ChIn(1) = Esc then Exit;		\wait for keystroke

\Solve maze using Depth-First Search
for Y:= 0 to Rows do
    for X:= 0 to Cols do
	Visited(X, Y):= false; 		\no cells visited
Solved:= false;
DFS(StartX, StartY);
ChOut(0, Bel); ChOut(0, Bel);		\ta-da!
if ChIn(1) = Esc then Exit;		\wait for keystroke

\Solve maze using Breadth-First Search
SetVid($13);				\320x200x8 graphics
RanSeed(Seed);				\generate same maze as before
GenMaze;
ShowMaze;
if ChIn(1) = Esc then Exit;		\wait for keystroke

for Y:= 0 to Rows do
    for X:= 0 to Cols do
	Visited(X, Y):= false; 		\no cells visited
BFS(StartX, StartY);
ChOut(0, Bel); ChOut(0, Bel);		\ta-da!
if ChIn(1) = Esc then Exit;		\wait for keystroke

SetVid($13);				\320x200x8 graphics
RanSeed(Seed);				\generate same maze as before
GenMaze;
ShowMaze;
if ChIn(1) = Esc then Exit;		\wait for keystroke

SetVid($03);				\restore normal text mode
end;	\Main
